# -*- coding: iso-8859-1 -*-

###
#COPYRIGHT © 2013 THE OHIO STATE UNIVERSITY
#ALL RIGHTS RESERVED
#
#PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
#REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR NONCOMMERCIAL
#EDUCATION AND RESEARCH PURPOSES, SO LONG AS NO FEE IS CHARGED, AND SO
#LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION, AND THE
#DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS THE NAME OF
#THE OHIO STATE UNIVERSITY IS NOT USED IN ANY ADVERTISING OR PUBLICITY
#PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE WITHOUT
#SPECIFIC, WRITTEN PRIOR AUTHORIZATION.
#
#THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE OHIO
#STATE UNIVERSITY AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
#WARRANTY BY THE OHIO STATE UNIVERSITY OF ANY KIND, EITHER EXPRESS OR
#IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
#MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE OHIO STATE
#UNIVERSITY SHALL NOT BE LIABLE FOR ANY DAMAGES, INCLUDING SPECIAL,
#INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY
#CLAIM ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE,
#EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
#DAMAGES.
###

try:
    import exceptions
    import cPickle as pickle
    import sqlite3
    import csv
    import gzip
    import threading
    import time
    import argparse
except (ImportError),err: raise

class PrEMeR_CG():
    """
    CpG Binner.

    Creates cgb files from cpg index, sam, and fragment length distribution files.
    """

    #=============================================
    # NOTES
    #=============================================

    #=============================================
    # GLOBAL CLASS VARIABLES
    #=============================================

    # holds the index of cpg dinucleotides in genome
    __cgindex = None

    # holds the cdf of the fragments to be binned
    __cdf = None

    # holds the sam generator
    __sam = None

    # sam header selected
    __samtype = None

    # header fields for SAM file types
    __samheaders = {
        'sam':['QNAME','FLAG','RNAME','POS','MAPQ','CIGAR','RNEXT','PNEXT','TLEN','SEQ','QUAL'],
        'tophat':['QNAME','FLAG','RNAME','POS','MAPQ','CIGAR','RNEXT','PNEXT','TLEN','SEQ','QUAL'],
        'bismark':['QNAME','FLAG','RNAME','POS','MAPQ','CIGAR','RNEXT','PNEXT','TLEN','SEQ','QUAL','NM-TAG','XX-TAG','XM-TAG','XR-TAG','XG-TAG'],
    }

    # thread locks
    __lock = threading.RLock()

    #=============================================
    # CLASS SPECIFIC EXCEPTIONS
    #=============================================

    class InitError(exceptions.Exception):
        """
        Class initialisation error.
        """

        def __init__(self,value):
            self.value = value
            return

        def __str__(self):
            return repr(self.value)

    class ParsingError(exceptions.Exception):
        """
        Parsing error.
        """

        def __init__(self,value):
            self.value = value
            return

        def __str__(self):
            return repr(self.value)

    #=============================================
    # CLASS METHODS
    #=============================================

    def __init__(self):
        """
        Constructor. Instantiates the cg binner.
        """
        return

    def bin(self,fout,textout,nthreads=1): #,cgindex=None,sam=None,cdf=None,ada=0,samtype='sam',nthreads=1):
        """
        Binner. Bins reads into individual cg loci.

        Parameters:
            fout - write opened file handle or file name string of desired output file
            sam - read opened file handle or file name string of target sam input file
            cdf - read opened file handle or file name string of target cdf input file
            ada - adapter length to trim from cdf
        """

        # try to open the output files
        try:
            if textout: gzipfh = gzip.open(fout+'.cgb.gz','w')
	    pickfh = open(fout+'.cgb.pkl','wb')
        except (IOError),err: raise IOError('Unable to open file for output')

        # spawn worker threads
        for i in xrange(nthreads):
            try: threads.append(threading.Thread(target=self.__worker_binner))
            except (NameError),err: threads = [threading.Thread(target=self.__worker_binner)]
            threads[-1].start()
        for thread in threads: thread.join()

        
        # normalize to cpm
        tot = 0.0
        for c in self.__cgindex:
            for p in self.__cgindex[c]:
                tot += self.__cgindex[c][p]
        if textout:
            print 'Outputing to text file'
	    # write cgindex to file
            gzipfh.write('\t'.join(['#chr','start','cov','cpm'])+'\n')
            for c in sorted(self.__cgindex):
            	for p in sorted(self.__cgindex[c]):
               		cpm = self.__cgindex[c][p]/tot
               		gzipfh.write('\t'.join([c,str(p),str(self.__cgindex[c][p]),str(cpm)])+'\n')
        pickle.dump(self.__cgindex,pickfh)
        # done binning
        return

    def loadcgindex(self,cgindex):
        """
        Instantiates the cg binner for a specific reference genome.

        Parameters:
            cgindex - read opened file handle or file name string of target cg index input file (pickle)
        """

        # try loading the cg index
        try: 
	    cgfh = open(cgindex, 'rb')
	    self.__cgindex = pickle.load(cgindex)
	except (TypeError),err:
            if type(cgindex) is file or type(cgindex) is gzip.GzipFile: raise self.InitError('unable to load pickle file')
            try: # it's not a known open file handle, so try opening it
                fh = gzip.open(cgindex,'r')
                fh.read()
                fh.seek(0)
            except (IOError),err: # does not appear to be gzipped
                try: fh = open(cgindex,'r')
                except (IOError),err: raise self.InitError('unable to open file for reading')
            # now try loading the cg index again
            try: self.__cgindex = pickle.load(fh)
            except (TypeError),err: raise self.InitError('unable to load pickle file')

        # we're done initializing

        return

    def loadcdf(self,cdf,ada=0):
        """
        CDF parser. Converts a cdf file to a cdf list.

        Parameters:
            cdf - read opened file handle or file name string of target cdf input file
            ada - adapter length to trim from cdf

        Returns:
            list of floating point values representing the fragment size cdf post adapter trimming
            position in the list corresponds to the position from the site (0 inclusive)
        """

        # try to open the file
        try: # see if cdf is a string and can be opened
            fh = gzip.open(cdf,'r')
            fh.read()
            fh.seek(0)
        except (IOError),err: # it's not a gzip
            try: fh = open(cdf,'r')
            except (IOError),err: raise PrEMeR_CG.ParsingError('cannot open file for parsing')
        except (TypeError),err: fh = cdf # assume it's an open file for now

        # try to get the position in the file and sample the file
        try: fh_start = fh.tell()
        except (AttributeError),err: raise PrEMeR_CG.ParsingError('parameter does not appear to be a file')
        fh_sample = fh.read(1024)
        fh.seek(fh_start)

        # dynamically determine the structure of the file
        format = csv.Sniffer().sniff(fh_sample)
        reader = csv.DictReader(fh,dialect=format)
        if not csv.Sniffer().has_header(fh_sample): raise PrEMeR_CG.ParsingError('file does not have a header')
        elif reader.fieldnames[0][0] is '#': reader.fieldnames[0] = reader.fieldnames[0][1:]

        # parse lines of the cdf file
        try:
            self.__cdf = []
            for lcnt,line in enumerate(fh):
                flds = line.strip('\n').split('\t')
                try:
                    if int(flds[reader.fieldnames.index('pos')])-ada < 0: continue
                    self.__cdf.append(float(flds[reader.fieldnames.index('cdf')]))
                except (TypeError),err: raise PrEMeR_CG.ParsingError('non-float value on line '+str(lcnt))
        except (AttributeError),err: raise PrEMeR_CG.ParsingError('parameter does not appear to be a file')
        fh.seek(fh_start)

        # return the cdf list
        return

    def loadsam(self,sam,samtype='sam'):
        """
        SAM parser. Produces a generator of pertinent sam file information.

        Parameters:
            sam - read opened file handle or file name string of target sam input file

        Returns:
            generator of dictionaries of sam file information
        """

        # try to open the file
        try: # see if sam is a string and can be opened
            fh = gzip.open(sam,'r')
            fh.read()
            fh.seek(0)
        except (IOError),err: # it's not a gzip
            try: fh = open(sam,'r')
            except (IOError),err: raise PrEMeR_CG.ParsingError('cannot open file for parsing')
        except (TypeError),err: fh = sam # assume it's an open file for now

        # bypass all the header information that we don't care about
        try: fh_truestart = fh.tell()
        except (AttributeError),err: raise PrEMeR_CG.ParsingError('file parameter does not have .tell()')
        while True:
            try: line = fh.readline()
            except (AttributeError),err: raise PrEMeR_CG.ParsingError('file parameter does not have .readline()')
            if line[0] is '@': fh_start = fh.tell()
            else:
                try: fh.seek(fh_start) # mark the beginning of pertinent information
                except (NameError),err:
                    fh_start = fh_truestart
                    fh.seek(fh_start)
                break
        fh_sample = '\n'.join(fh.readlines(1024))
        fh.seek(fh_start)
        
        # dynamically determine the structure of the file
        format = csv.Sniffer().sniff(fh_sample)
        reader = csv.DictReader(fh,dialect=format)
        if not csv.Sniffer().has_header(fh_sample): reader.fieldnames = self.__samheaders[samtype]
        elif reader.fieldnames[0][0] is '#': reader.fieldnames[0] = reader.fieldnames[0][1:]
        else: raise PrEMeR_CG.ParsingError('file appears to have custom header - does not begin with #')

        self.__samtype = samtype
        self.__sam = PrEMeR_CG.__threaded_iterator(reader)

        # we're done parsing the sam file
        return
        
    #=============================================
    # THREADING STUFF
    #=============================================

    # thread safe wrapper for iterators and generators
    class __threaded_iterator:
        def __init__(self,itt):
            self.lock = threading.Lock() # <-- this makes it safe
            self.itt = itt.__iter__()
            return

        def __iter__(self):
            return self

        def next(self):
            self.lock.acquire()
            try: return self.itt.next()
            finally: self.lock.release()

    # worker thread for binning sam lines into cg loci
    def __worker_binner(self):
        # consume parts of the sam file
        for line in self.__sam:
            try: lcnt += 1
            except (NameError),err: lcnt = 1
            chro = line['RNAME'] # get chromosome
            if chro not in self.__cgindex: continue

            try: stra = bool(int(line['FLAG']) & 16) # get strand information from mask
            except (TypeError),err: raise PrEMeR_CG.ParsingError('FLAG field not int')
            else: stra = (-1)**stra # convert to -1 or +1

            try: star = int(line['POS']) # get start
            except (TypeError),err: raise PrEMeR_CG.ParsingError('POS field not int')
            if stra == -1: star += len(line['SEQ'])-1 # if reverse strand, go to the end

            # intersect with the cgindex
            if not self.__cdf: cdf = [1.0]*len(line['SEQ'])
            else: cdf = self.__cdf
            positions = []
            values = []
            for pos,val in enumerate(cdf):
                if star < 0: break
                if star in self.__cgindex[chro]:
                    positions.append(star)
                    values.append(val)
                star += stra
            for pos,val in zip(positions,values):
                self.__lock.acquire()
                self.__cgindex[chro][pos] += val/sum(values)
                self.__lock.release()

        # done processing
        
        return
